#
#  mpq_ffi.rb
#  DbcExtractor
# FFI version of the mpq extractor, you need to have the stormlib.dll/so/dylib in your LD_LIBRARY_PATH or in the current directory
#

require 'fileutils'
require 'ffi'

class Entry
  include Comparable
  attr_accessor :path, :os_path
  def initialize(p,s)
    @path = p
    @os_path = s
  end
  def <=>(other)
    @path <=> other.path
  end
end

#Entry = Struct.new(:path,:os_path)

module StormLib
  extend FFI::Library
  SFILE_OPEN_PATCHED_FILE = 0x00000001
  SFILE_OPEN_FROM_MPQ  = 0x00000000
  SFILE_VERIFY_ALL = 0x0000000F
  VERIFY_OPEN_ERROR = 0x0001
  VERIFY_READ_ERROR = 0x0002 
  VERIFY_FILE_HAS_SECTOR_CRC = 0x0004
  VERIFY_FILE_SECTOR_CRC_ERROR = 0x0008
  VERIFY_FILE_HAS_CHECKSUM  = 0x0010
  VERIFY_FILE_CHECKSUM_ERROR = 0x0020
  VERIFY_FILE_HAS_MD5 = 0x0040
  VERIFY_FILE_MD5_ERROR = 0x0080
  VERIFY_FILE_HAS_RAW_MD5 = 0x0100
  VERIFY_FILE_RAW_MD5_ERROR = 0x0200
  VERIFY_FILE_ERROR_MASK = (VERIFY_OPEN_ERROR | VERIFY_READ_ERROR | VERIFY_FILE_SECTOR_CRC_ERROR | VERIFY_FILE_CHECKSUM_ERROR | VERIFY_FILE_MD5_ERROR | VERIFY_FILE_RAW_MD5_ERROR)
  ffi_lib 'Storm'
  attach_function :SFileOpenArchive, [:string, :int, :int, :pointer], :int
  attach_function :GetLastError, [], :int
  attach_function :SFileCloseArchive, [:pointer], :bool
  attach_function :SFileFindFirstFile, [:pointer, :string, :pointer, :string], :pointer
  attach_function :SFileFindNextFile, [:pointer, :pointer] ,:bool
  attach_function :SFileOpenPatchArchive, [:pointer,:string,:string,:int], :bool
  attach_function :SFileVerifyFile, [:pointer,:string,:int], :int
  attach_function :SFileExtractFile, [:pointer,:string,:string,:int], :bool
  class FileFindData < FFI::Struct
      layout :cFileName, [:char,1024],
        :szPlainName, :string,
        :hashIndex, :int,
        :blockIndex, :int,
        :fileSize, :int,
        :fileFlags, :int,
        :compSize, :int,
        :fileTimeLo, :int,
        :filetimeHi, :int,
        :locale, :int
    end
end

class MPQ
	def initialize
		@mph = FFI::MemoryPointer.new(:pointer, 1)
	end
	
	def MPQ.valid_dir?(path)
		Dir.exists?("#{path}/Data") and Dir.exists?("#{path}/World of Warcraft.app")
	end
	
	def open(file)
		puts "Opening #{file}"
		raise "Unable to open archive #{StormLib.GetLastError()}" unless StormLib.SFileOpenArchive(file,0,0x0100,@mph)
		true
	end
	def close()
		StormLib.SFileCloseArchive(@mph.read_pointer)
		true
	end
	def list(pattern)
		h = []
		x = StormLib::FileFindData.new
		fh = StormLib.SFileFindFirstFile(@mph.read_pointer,pattern,x.pointer,nil)
		if (fh)
			path = x[:cFileName].to_s
			os_path = path.gsub(/\\/,'/')
			entry = Entry.new(path,os_path)
			h << entry
			x = StormLib::FileFindData.new
			while StormLib.SFileFindNextFile(fh,x.pointer)
				path = x[:cFileName].to_s
				os_path = path.gsub(/\\/,'/')
				entry = Entry.new(path,os_path)
				h << entry
				#x = StormLib::FileFindData.new
			end
		end
		h
	end
	def patch(file,prefix)
		puts "Patching #{file}"
		raise "Unable to open archive" unless StormLib.SFileOpenPatchArchive(@mph.read_pointer,file,nil,0)
		true
	end
	def extract(entry,dest)
		os_file = "#{dest}/#{entry.os_path}"
		unless File.exists?(os_file)
			FileUtils.makedirs(File.dirname(os_file))
		else
			File.delete(os_file)
		end
		rc = StormLib.SFileVerifyFile(@mph.read_pointer,entry.path,StormLib::SFILE_VERIFY_ALL)
		if (rc & StormLib::VERIFY_FILE_ERROR_MASK) == 0
			@cb.call("Creating #{os_file}") if @cb
			raise "Failed to extract #{os_file} #{GetLastError()}" unless StormLib.SFileExtractFile(@mph.read_pointer,entry.path,os_file,StormLib::SFILE_OPEN_PATCHED_FILE)
		else
			raise "Failed to verify file #{entry.path} #{rc & StormLib::VERIFY_FILE_ERROR_MASK} #{rc} #{StromLib.GetLastError()}"
		end
	end
end

class Warcraft
	def initialize(path)
		@wow_path = "#{path}/Data"
	end
	
	def add_callback(&callback)
		@cb = callback
	end
	
	def load_mpqs(locale="enUS")
		mpq = MPQ.new
		mpq.open("#{@wow_path}/#{locale}/locale-#{locale}.MPQ")
		mpq.patch("#{@wow_path}/misc.MPQ",locale)
		# now add the 
		base_files = Dir.entries("#{@wow_path}")
		base_files.delete_if do |item| not item =~ /wow.*MPQ/ end
		base_files.sort.each do |b|
			x = b.gsub("base",locale)
			@cb.call "#{@wow_path}/#{b}" if @cb
			mpq.patch("#{@wow_path}/#{b}","enUS")
			if File.exists?("#{@wow_path}/#{locale}/#{x}")
				mpq.patch("#{@wow_path}/#{locale}/#{x}",locale)
			end
		end		
		return mpq
	end
	
end
if __FILE__ == $0
	wow = Warcraft.new(ARGV[0])
	m = wow.load_mpqs()
	lst = m.list("*.dbc")
	lst.sort.each do |l|
		puts "Extracting #{l.os_path}"
		m.extract(l,ARGV[1])
	end
	m.close
end